@@ -20,9 +20,12 @@ from .. import mlog
 from .. import compilers
 import json
 import subprocess
-from ..mesonlib import MesonException, get_compiler_for_source, classify_unity_sources
+from ..mesonlib import MesonException, get_meson_script
+from ..mesonlib import get_compiler_for_source, classify_unity_sources
+from ..compilers import CompilerArgs
+from collections import OrderedDict
 
-class CleanTrees():
+class CleanTrees:
     '''
     Directories outputted by custom targets that have to be manually cleaned
     because on Linux `ninja clean` only deletes empty directories.
@@ -31,11 +34,12 @@ class CleanTrees():
         self.build_dir = build_dir
         self.trees = trees
 
-class InstallData():
-    def __init__(self, source_dir, build_dir, prefix):
+class InstallData:
+    def __init__(self, source_dir, build_dir, prefix, strip_bin, mesonintrospect):
         self.source_dir = source_dir
-        self.build_dir= build_dir
+        self.build_dir = build_dir
         self.prefix = prefix
+        self.strip_bin = strip_bin
         self.targets = []
         self.headers = []
         self.man = []
@@ -44,8 +48,9 @@ class InstallData():
         self.po = []
         self.install_scripts = []
         self.install_subdirs = []
+        self.mesonintrospect = mesonintrospect
 
-class ExecutableSerialisation():
+class ExecutableSerialisation:
     def __init__(self, name, fname, cmd_args, env, is_cross, exe_wrapper,
                  workdir, extra_paths, capture):
         self.name = name
@@ -74,9 +79,27 @@ class TestSerialisation:
         self.workdir = workdir
         self.extra_paths = extra_paths
 
+class OptionProxy:
+    def __init__(self, name, value):
+        self.name = name
+        self.value = value
+
+class OptionOverrideProxy:
+    '''Mimic an option list but transparently override
+    selected option values.'''
+    def __init__(self, overrides, options):
+        self.overrides = overrides
+        self.options = options
+
+    def __getitem__(self, option_name):
+        base_opt = self.options[option_name]
+        if option_name in self.overrides:
+            return OptionProxy(base_opt.name, base_opt.validate_value(self.overrides[option_name]))
+        return base_opt
+
 # This class contains the basic functionality that is needed by all backends.
 # Feel free to move stuff in and out of it as you see fit.
-class Backend():
+class Backend:
     def __init__(self, build):
         self.build = build
         self.environment = build.environment
@@ -90,7 +113,7 @@ class Backend():
     def get_target_filename(self, t):
         if isinstance(t, build.CustomTarget):
             if len(t.get_outputs()) != 1:
-                mlog.warning('custom_target {!r} has more than one output! ' \
+                mlog.warning('custom_target {!r} has more than one output! '
                              'Using the first one.'.format(t.name))
             filename = t.get_outputs()[0]
         else:
@@ -101,6 +124,12 @@ class Backend():
     def get_target_filename_abs(self, target):
         return os.path.join(self.environment.get_build_dir(), self.get_target_filename(target))
 
+    def get_option_for_target(self, option_name, target):
+        if option_name in target.option_overrides:
+            override = target.option_overrides[option_name]
+            return self.environment.coredata.validate_option_value(option_name, override)
+        return self.environment.coredata.get_builtin_option(option_name)
+
     def get_target_filename_for_linking(self, target):
         # On some platforms (msvc for instance), the file that is used for
         # dynamic linking is not the same as the dynamic library itself. This
@@ -113,12 +142,6 @@ class Backend():
             return os.path.join(self.get_target_dir(target), target.get_filename())
         raise AssertionError('BUG: Tried to link to something that\'s not a library')
 
-    def get_target_debug_filename(self, target):
-        fname = target.get_debug_filename()
-        if not fname:
-            raise AssertionError("BUG: Tried to generate debug filename when it doesn't exist")
-        return os.path.join(self.get_target_dir(target), fname)
-
     def get_target_dir(self, target):
         if self.environment.coredata.get_builtin_option('layout') == 'mirror':
             dirname = target.get_subdir()
@@ -126,6 +149,10 @@ class Backend():
             dirname = 'meson-out'
         return dirname
 
+    def get_target_source_dir(self, target):
+        dirname = os.path.join(self.build_to_src, self.get_target_dir(target))
+        return dirname
+
     def get_target_private_dir(self, target):
         dirname = os.path.join(self.get_target_dir(target), target.get_basename() + target.type_suffix())
         return dirname
@@ -156,8 +183,10 @@ class Backend():
         compsrcs = classify_unity_sources(target.compilers.values(), unity_src)
 
         def init_language_file(suffix):
-            outfilename = os.path.join(self.get_target_private_dir_abs(target),
-                                       self.get_unity_source_filename(target, suffix))
+            unity_src_name = self.get_unity_source_filename(target, suffix)
+            unity_src_subdir = self.get_target_private_dir_abs(target)
+            outfilename = os.path.join(unity_src_subdir,
+                                       unity_src_name)
             outfileabs = os.path.join(self.environment.get_build_dir(),
                                       outfilename)
             outfileabs_tmp = outfileabs + '.tmp'
@@ -165,12 +194,11 @@ class Backend():
             outfileabs_tmp_dir = os.path.dirname(outfileabs_tmp)
             if not os.path.exists(outfileabs_tmp_dir):
                 os.makedirs(outfileabs_tmp_dir)
-            result.append(outfilename)
+            result.append(mesonlib.File(True, unity_src_subdir, unity_src_name))
             return open(outfileabs_tmp, 'w')
 
         # For each language, generate a unity source file and return the list
         for comp, srcs in compsrcs.items():
-            lang = comp.get_language()
             with init_language_file(comp.get_default_suffix()) as ofile:
                 for src in srcs:
                     ofile.write('#include<%s>\n' % src)
@@ -178,7 +206,7 @@ class Backend():
         return result
 
     def relpath(self, todir, fromdir):
-        return os.path.relpath(os.path.join('dummyprefixdir', todir),\
+        return os.path.relpath(os.path.join('dummyprefixdir', todir),
                                os.path.join('dummyprefixdir', fromdir))
 
     def flatten_object_list(self, target, proj_dir_to_build_root=''):
@@ -191,32 +219,37 @@ class Backend():
             elif isinstance(obj, mesonlib.File):
                 obj_list.append(obj.rel_to_builddir(self.build_to_src))
             elif isinstance(obj, build.ExtractedObjects):
-                obj_list += self.determine_ext_objs(obj, proj_dir_to_build_root)
+                obj_list += self.determine_ext_objs(target, obj, proj_dir_to_build_root)
             else:
                 raise MesonException('Unknown data type in object list.')
         return obj_list
 
-    def serialise_executable(self, exe, cmd_args, workdir, env={},
+    def serialize_executable(self, exe, cmd_args, workdir, env={},
                              capture=None):
-        import uuid
+        import hashlib
         # Can't just use exe.name here; it will likely be run more than once
         if isinstance(exe, (dependencies.ExternalProgram,
                             build.BuildTarget, build.CustomTarget)):
             basename = exe.name
         else:
             basename = os.path.basename(exe)
-        scratch_file = 'meson_exe_{0}_{1}.dat'.format(basename,
-                                                      str(uuid.uuid4())[:8])
+        # Take a digest of the cmd args, env, workdir, and capture. This avoids
+        # collisions and also makes the name deterministic over regenerations
+        # which avoids a rebuild by Ninja because the cmdline stays the same.
+        data = bytes(str(sorted(env.items())) + str(cmd_args) + str(workdir) + str(capture),
+                     encoding='utf-8')
+        digest = hashlib.sha1(data).hexdigest()
+        scratch_file = 'meson_exe_{0}_{1}.dat'.format(basename, digest)
         exe_data = os.path.join(self.environment.get_scratch_dir(), scratch_file)
         with open(exe_data, 'wb') as f:
             if isinstance(exe, dependencies.ExternalProgram):
-                exe_fullpath = exe.fullpath
+                exe_cmd = exe.get_command()
                 exe_needs_wrapper = False
             elif isinstance(exe, (build.BuildTarget, build.CustomTarget)):
-                exe_fullpath = [self.get_target_filename_abs(exe)]
+                exe_cmd = [self.get_target_filename_abs(exe)]
                 exe_needs_wrapper = exe.is_cross
             else:
-                exe_fullpath = [exe]
+                exe_cmd = [exe]
                 exe_needs_wrapper = False
             is_cross = exe_needs_wrapper and \
                 self.environment.is_cross_build() and \
@@ -226,24 +259,24 @@ class Backend():
                 exe_wrapper = self.environment.cross_info.config['binaries'].get('exe_wrapper', None)
             else:
                 exe_wrapper = None
-            if mesonlib.is_windows():
+            if mesonlib.is_windows() or mesonlib.is_cygwin():
                 extra_paths = self.determine_windows_extra_paths(exe)
             else:
                 extra_paths = []
-            es = ExecutableSerialisation(basename, exe_fullpath, cmd_args, env,
+            es = ExecutableSerialisation(basename, exe_cmd, cmd_args, env,
                                          is_cross, exe_wrapper, workdir,
                                          extra_paths, capture)
             pickle.dump(es, f)
         return exe_data
 
-    def serialise_tests(self):
+    def serialize_tests(self):
         test_data = os.path.join(self.environment.get_scratch_dir(), 'meson_test_setup.dat')
         with open(test_data, 'wb') as datafile:
             self.write_test_file(datafile)
         benchmark_data = os.path.join(self.environment.get_scratch_dir(), 'meson_benchmark_setup.dat')
         with open(benchmark_data, 'wb') as datafile:
             self.write_benchmark_file(datafile)
-        return (test_data, benchmark_data)
+        return test_data, benchmark_data
 
     def determine_linker(self, target):
         '''
@@ -261,32 +294,45 @@ class Backend():
             raise MesonException(m.format(target.name))
         return l
 
-    def object_filename_from_source(self, target, source):
+    def determine_rpath_dirs(self, target):
+        link_deps = target.get_all_link_deps()
+        result = []
+        for ld in link_deps:
+            prospective = self.get_target_dir(ld)
+            if prospective not in result:
+                result.append(prospective)
+        return result
+
+    def object_filename_from_source(self, target, source, is_unity):
         if isinstance(source, mesonlib.File):
             source = source.fname
         # foo.vala files compile down to foo.c and then foo.c.o, not foo.vala.o
         if source.endswith('.vala'):
+            if is_unity:
+                return source[:-5] + '.c.' + self.environment.get_object_suffix()
             source = os.path.join(self.get_target_private_dir(target), source[:-5] + '.c')
         return source.replace('/', '_').replace('\\', '_') + '.' + self.environment.get_object_suffix()
 
-    def determine_ext_objs(self, extobj, proj_dir_to_build_root):
+    def determine_ext_objs(self, target, extobj, proj_dir_to_build_root):
         result = []
         targetdir = self.get_target_private_dir(extobj.target)
         # With unity builds, there's just one object that contains all the
         # sources, and we only support extracting all the objects in this mode,
         # so just return that.
-        if self.environment.coredata.get_builtin_option('unity'):
+        if self.is_unity(target):
             comp = get_compiler_for_source(extobj.target.compilers.values(),
                                            extobj.srclist[0])
-            # The unity object name uses the full absolute path of the source file
-            osrc = os.path.join(self.get_target_private_dir_abs(extobj.target),
-                                self.get_unity_source_filename(extobj.target,
-                                                               comp.get_default_suffix()))
-            objname = self.object_filename_from_source(extobj.target, osrc)
+            # There is a potential conflict here, but it is unlikely that
+            # anyone both enables unity builds and has a file called foo-unity.cpp.
+            osrc = self.get_unity_source_filename(extobj.target,
+                                                  comp.get_default_suffix())
+            osrc = os.path.join(self.get_target_private_dir(extobj.target), osrc)
+            objname = self.object_filename_from_source(extobj.target, osrc, True)
+            objname = objname.replace('/', '_').replace('\\', '_')
             objpath = os.path.join(proj_dir_to_build_root, targetdir, objname)
             return [objpath]
         for osrc in extobj.srclist:
-            objname = self.object_filename_from_source(extobj.target, osrc)
+            objname = self.object_filename_from_source(extobj.target, osrc, False)
             objpath = os.path.join(proj_dir_to_build_root, targetdir, objname)
             result.append(objpath)
         return result
@@ -297,7 +343,7 @@ class Backend():
         includeargs = compiler.get_include_args(pchpath, False)
         for lang in ['c', 'cpp']:
             p = target.get_pch(lang)
-            if len(p) == 0:
+            if not p:
                 continue
             if compiler.can_compile(p[-1]):
                 header = p[0]
@@ -334,32 +380,75 @@ class Backend():
         return extra_args
 
     def generate_basic_compiler_args(self, target, compiler, no_warn_args=False):
-        commands = []
+        # Create an empty commands list, and start adding arguments from
+        # various sources in the order in which they must override each other
+        # starting from hard-coded defaults followed by build options and so on.
+        commands = CompilerArgs(compiler)
+
+        copt_proxy = OptionOverrideProxy(target.option_overrides, self.environment.coredata.compiler_options)
+        # First, the trivial ones that are impossible to override.
+        #
+        # Add -nostdinc/-nostdinc++ if needed; can't be overriden
         commands += self.get_cross_stdlib_args(target, compiler)
+        # Add things like /NOLOGO or -pipe; usually can't be overriden
         commands += compiler.get_always_args()
+        # Only add warning-flags by default if the buildtype enables it, and if
+        # we weren't explicitly asked to not emit warnings (for Vala, f.ex)
         if no_warn_args:
             commands += compiler.get_no_warn_args()
-        else:
-            commands += compiler.get_warn_args(self.environment.coredata.get_builtin_option('warning_level'))
-        commands += compiler.get_option_compile_args(self.environment.coredata.compiler_options)
-        commands += self.build.get_global_args(compiler)
-        commands += self.build.get_project_args(compiler, target.subproject)
-        commands += self.environment.coredata.external_args[compiler.get_language()]
-        commands += self.escape_extra_args(compiler, target.get_extra_args(compiler.get_language()))
-        commands += compiler.get_buildtype_args(self.environment.coredata.get_builtin_option('buildtype'))
-        if self.environment.coredata.get_builtin_option('werror'):
+        elif self.get_option_for_target('buildtype', target) != 'plain':
+            commands += compiler.get_warn_args(self.get_option_for_target('warning_level', target))
+        # Add -Werror if werror=true is set in the build options set on the
+        # command-line or default_options inside project(). This only sets the
+        # action to be done for warnings if/when they are emitted, so it's ok
+        # to set it after get_no_warn_args() or get_warn_args().
+        if self.get_option_for_target('werror', target):
             commands += compiler.get_werror_args()
+        # Add compile args for c_* or cpp_* build options set on the
+        # command-line or default_options inside project().
+        commands += compiler.get_option_compile_args(copt_proxy)
+        # Add buildtype args: optimization level, debugging, etc.
+        commands += compiler.get_buildtype_args(self.get_option_for_target('buildtype', target))
+        # Add compile args added using add_project_arguments()
+        commands += self.build.get_project_args(compiler, target.subproject)
+        # Add compile args added using add_global_arguments()
+        # These override per-project arguments
+        commands += self.build.get_global_args(compiler)
+        if not target.is_cross:
+            # Compile args added from the env: CFLAGS/CXXFLAGS, etc. We want these
+            # to override all the defaults, but not the per-target compile args.
+            commands += self.environment.coredata.external_args[compiler.get_language()]
+        # Always set -fPIC for shared libraries
         if isinstance(target, build.SharedLibrary):
             commands += compiler.get_pic_args()
+        # Set -fPIC for static libraries by default unless explicitly disabled
         if isinstance(target, build.StaticLibrary) and target.pic:
             commands += compiler.get_pic_args()
-        for dep in target.get_external_deps():
-            # Cflags required by external deps might have UNIX-specific flags,
-            # so filter them out if needed
-            commands += compiler.unix_compile_flags_to_native(dep.get_compile_args())
+        # Add compile args needed to find external dependencies. Link args are
+        # added while generating the link command.
+        # NOTE: We must preserve the order in which external deps are
+        # specified, so we reverse the list before iterating over it.
+        for dep in reversed(target.get_external_deps()):
+            if compiler.language == 'vala':
+                if isinstance(dep, dependencies.PkgConfigDependency):
+                    if dep.name == 'glib-2.0' and dep.version_reqs is not None:
+                        for req in dep.version_reqs:
+                            if req.startswith(('>=', '==')):
+                                commands += ['--target-glib', req[2:]]
+                                break
+                    commands += ['--pkg', dep.name]
+                elif isinstance(dep, dependencies.ExternalLibrary):
+                    commands += dep.get_link_args('vala')
+            else:
+                commands += dep.get_compile_args()
+            # Qt needs -fPIC for executables
+            # XXX: We should move to -fPIC for all executables
             if isinstance(target, build.Executable):
-                commands += dep.get_exe_args()
-
+                commands += dep.get_exe_args(compiler)
+            # For 'automagic' deps: Boost and GTest. Also dependency('threads').
+            # pkg-config puts the thread flags itself via `Cflags:`
+            if dep.need_threads():
+                commands += compiler.thread_flags()
         # Fortran requires extra include directives.
         if compiler.language == 'fortran':
             for lt in target.link_targets:
@@ -373,15 +462,11 @@ class Backend():
         for d in deps:
             if not isinstance(d, (build.StaticLibrary, build.SharedLibrary)):
                 raise RuntimeError('Tried to link with a non-library target "%s".' % d.get_basename())
-            if isinstance(compiler, compilers.LLVMDCompiler):
-                args.extend(['-L', self.get_target_filename_for_linking(d)])
+            if isinstance(compiler, (compilers.LLVMDCompiler, compilers.DmdDCompiler)):
+                d_arg = '-L' + self.get_target_filename_for_linking(d)
             else:
-                args.append(self.get_target_filename_for_linking(d))
-            # If you have executable e that links to shared lib s1 that links to shared library s2
-            # you have to specify s2 as well as s1 when linking e even if e does not directly use
-            # s2. Gcc handles this case fine but Clang does not for some reason. Thus we need to
-            # explictly specify all libraries every time.
-            args += self.build_target_link_arguments(compiler, d.get_dependencies())
+                d_arg = self.get_target_filename_for_linking(d)
+            args.append(d_arg)
         return args
 
     def determine_windows_extra_paths(self, target):
@@ -412,9 +497,9 @@ class Backend():
         for t in tests:
             exe = t.get_exe()
             if isinstance(exe, dependencies.ExternalProgram):
-                fname = exe.fullpath
+                cmd = exe.get_command()
             else:
-                fname = [os.path.join(self.environment.get_build_dir(), self.get_target_filename(t.get_exe()))]
+                cmd = [os.path.join(self.environment.get_build_dir(), self.get_target_filename(t.get_exe()))]
             is_cross = self.environment.is_cross_build() and \
                 self.environment.cross_info.need_cross_compiler() and \
                 self.environment.cross_info.need_exe_wrapper()
@@ -422,38 +507,44 @@ class Backend():
                 exe_wrapper = self.environment.cross_info.config['binaries'].get('exe_wrapper', None)
             else:
                 exe_wrapper = None
-            if mesonlib.is_windows():
+            if mesonlib.is_windows() or mesonlib.is_cygwin():
                 extra_paths = self.determine_windows_extra_paths(exe)
             else:
                 extra_paths = []
             cmd_args = []
             for a in t.cmd_args:
+                if hasattr(a, 'held_object'):
+                    a = a.held_object
                 if isinstance(a, mesonlib.File):
                     a = os.path.join(self.environment.get_build_dir(), a.rel_to_builddir(self.build_to_src))
-                cmd_args.append(a)
-            ts = TestSerialisation(t.get_name(), t.suite, fname, is_cross, exe_wrapper,
+                    cmd_args.append(a)
+                elif isinstance(a, str):
+                    cmd_args.append(a)
+                elif isinstance(a, build.Target):
+                    cmd_args.append(self.get_target_filename(a))
+                else:
+                    raise MesonException('Bad object in test command.')
+            ts = TestSerialisation(t.get_name(), t.suite, cmd, is_cross, exe_wrapper,
                                    t.is_parallel, cmd_args, t.env, t.should_fail,
                                    t.timeout, t.workdir, extra_paths)
             arr.append(ts)
         pickle.dump(arr, datafile)
 
-
     def generate_depmf_install(self, d):
         if self.build.dep_manifest_name is None:
             return
         ifilename = os.path.join(self.environment.get_build_dir(), 'depmf.json')
         ofilename = os.path.join(self.environment.get_prefix(), self.build.dep_manifest_name)
-        mfobj = {'type': 'dependency manifest',
-                 'version': '1.0'}
-        mfobj['projects'] = self.build.dep_manifest
+        mfobj = {'type': 'dependency manifest', 'version': '1.0', 'projects': self.build.dep_manifest}
         with open(ifilename, 'w') as f:
             f.write(json.dumps(mfobj))
-        d.data.append([ifilename, ofilename])
+        # Copy file from, to, and with mode unchanged
+        d.data.append([ifilename, ofilename, None])
 
     def get_regen_filelist(self):
         '''List of all files whose alteration means that the build
         definition needs to be regenerated.'''
-        deps = [os.path.join(self.build_to_src, df) \
+        deps = [os.path.join(self.build_to_src, df)
                 for df in self.interpreter.get_build_def_files()]
         if self.environment.is_cross_build():
             deps.append(os.path.join(self.build_to_src,
@@ -504,16 +595,45 @@ class Backend():
             newargs.append(arg)
         return newargs
 
+    def get_build_by_default_targets(self):
+        result = OrderedDict()
+        # Get all build and custom targets that must be built by default
+        for name, t in self.build.get_targets().items():
+            if t.build_by_default or t.install or t.build_always:
+                result[name] = t
+        # Get all targets used as test executables and arguments. These must
+        # also be built by default. XXX: Sometime in the future these should be
+        # built only before running tests.
+        for t in self.build.get_tests():
+            exe = t.exe
+            if hasattr(exe, 'held_object'):
+                exe = exe.held_object
+            if isinstance(exe, (build.CustomTarget, build.BuildTarget)):
+                result[exe.get_id()] = exe
+            for arg in t.cmd_args:
+                if hasattr(arg, 'held_object'):
+                    arg = arg.held_object
+                if not isinstance(arg, (build.CustomTarget, build.BuildTarget)):
+                    continue
+                result[arg.get_id()] = arg
+        return result
+
     def get_custom_target_provided_libraries(self, target):
         libs = []
         for t in target.get_generated_sources():
             if not isinstance(t, build.CustomTarget):
                 continue
-            for f in t.output:
+            for f in t.get_outputs():
                 if self.environment.is_library(f):
                     libs.append(os.path.join(self.get_target_dir(t), f))
         return libs
 
+    def is_unity(self, target):
+        optval = self.get_option_for_target('unity', target)
+        if optval == 'on' or (optval == 'subprojects' and target.subproject != ''):
+            return True
+        return False
+
     def get_custom_target_sources(self, target):
         '''
         Custom target sources can be of various object types; strings, File,
@@ -539,20 +659,37 @@ class Backend():
             srcs += fname
         return srcs
 
+    def get_custom_target_depend_files(self, target, absolute_paths=False):
+        deps = []
+        for i in target.depend_files:
+            if isinstance(i, mesonlib.File):
+                if absolute_paths:
+                    deps.append(i.absolute_path(self.environment.get_source_dir(),
+                                                self.environment.get_build_dir()))
+                else:
+                    deps.append(i.rel_to_builddir(self.build_to_src))
+            else:
+                if absolute_paths:
+                    deps.append(os.path.join(self.environment.get_build_dir(), i))
+                else:
+                    deps.append(os.path.join(self.build_to_src, i))
+        return deps
+
     def eval_custom_target_command(self, target, absolute_outputs=False):
-        # We only want the outputs to be absolute when using the VS backend
-        if not absolute_outputs:
-            ofilenames = [os.path.join(self.get_target_dir(target), i) for i in target.output]
-        else:
-            ofilenames = [os.path.join(self.environment.get_build_dir(), self.get_target_dir(target), i) \
-                          for i in target.output]
-        srcs = self.get_custom_target_sources(target)
+        # We want the outputs to be absolute only when using the VS backend
+        # XXX: Maybe allow the vs backend to use relative paths too?
+        source_root = self.build_to_src
+        build_root = '.'
         outdir = self.get_target_dir(target)
-        # Many external programs fail on empty arguments.
-        if outdir == '':
-            outdir = '.'
-        if target.absolute_paths:
+        if absolute_outputs:
+            source_root = self.environment.get_source_dir()
+            build_root = self.environment.get_source_dir()
             outdir = os.path.join(self.environment.get_build_dir(), outdir)
+        outputs = []
+        for i in target.get_outputs():
+            outputs.append(os.path.join(outdir, i))
+        inputs = self.get_custom_target_sources(target)
+        # Evaluate the command list
         cmd = []
         for i in target.command:
             if isinstance(i, build.Executable):
@@ -568,37 +705,14 @@ class Backend():
                 if target.absolute_paths:
                     i = os.path.join(self.environment.get_build_dir(), i)
             # FIXME: str types are blindly added ignoring 'target.absolute_paths'
+            # because we can't know if they refer to a file or just a string
             elif not isinstance(i, str):
                 err_msg = 'Argument {0} is of unknown type {1}'
                 raise RuntimeError(err_msg.format(str(i), str(type(i))))
-            for (j, src) in enumerate(srcs):
-                i = i.replace('@INPUT%d@' % j, src)
-            for (j, res) in enumerate(ofilenames):
-                i = i.replace('@OUTPUT%d@' % j, res)
-            if '@INPUT@' in i:
-                msg = 'Custom target {} has @INPUT@ in the command, but'.format(target.name)
-                if len(srcs) == 0:
-                    raise MesonException(msg + ' no input files')
-                if i == '@INPUT@':
-                    cmd += srcs
-                    continue
-                else:
-                    if len(srcs) > 1:
-                        raise MesonException(msg + ' more than one input file')
-                    i = i.replace('@INPUT@', srcs[0])
-            elif '@OUTPUT@' in i:
-                msg = 'Custom target {} has @OUTPUT@ in the command, but'.format(target.name)
-                if len(ofilenames) == 0:
-                    raise MesonException(msg + ' no output files')
-                if i == '@OUTPUT@':
-                    cmd += ofilenames
-                    continue
-                else:
-                    if len(ofilenames) > 1:
-                        raise MesonException(msg + ' more than one output file')
-                    i = i.replace('@OUTPUT@', ofilenames[0])
-            elif '@OUTDIR@' in i:
-                i = i.replace('@OUTDIR@', outdir)
+            elif '@SOURCE_ROOT@' in i:
+                i = i.replace('@SOURCE_ROOT@', source_root)
+            elif '@BUILD_ROOT@' in i:
+                i = i.replace('@BUILD_ROOT@', build_root)
             elif '@DEPFILE@' in i:
                 if target.depfile is None:
                     msg = 'Custom target {!r} has @DEPFILE@ but no depfile ' \
@@ -607,7 +721,7 @@ class Backend():
                 dfilename = os.path.join(outdir, target.depfile)
                 i = i.replace('@DEPFILE@', dfilename)
             elif '@PRIVATE_OUTDIR_' in i:
-                match = re.search('@PRIVATE_OUTDIR_(ABS_)?([^\/\s*]*)@', i)
+                match = re.search('@PRIVATE_OUTDIR_(ABS_)?([^/\s*]*)@', i)
                 if not match:
                     msg = 'Custom target {!r} has an invalid argument {!r}' \
                           ''.format(target.name, i)
@@ -617,10 +731,11 @@ class Backend():
                     lead_dir = ''
                 else:
                     lead_dir = self.environment.get_build_dir()
-                i = i.replace(source,
-                              os.path.join(lead_dir,
-                                           outdir))
+                i = i.replace(source, os.path.join(lead_dir, outdir))
             cmd.append(i)
+        # Substitute the rest of the template strings
+        values = mesonlib.get_filenames_templates_dict(inputs, outputs)
+        cmd = mesonlib.substitute_values(cmd, values)
         # This should not be necessary but removing it breaks
         # building GStreamer on Windows. The underlying issue
         # is problems with quoting backslashes on Windows
@@ -640,36 +755,15 @@ class Backend():
         #
         # https://github.com/mesonbuild/meson/pull/737
         cmd = [i.replace('\\', '/') for i in cmd]
-        return (srcs, ofilenames, cmd)
+        return inputs, outputs, cmd
 
     def run_postconf_scripts(self):
-        env = {'MESON_SOURCE_ROOT' : self.environment.get_source_dir(),
-               'MESON_BUILD_ROOT' : self.environment.get_build_dir()
-              }
+        env = {'MESON_SOURCE_ROOT': self.environment.get_source_dir(),
+               'MESON_BUILD_ROOT': self.environment.get_build_dir(),
+               'MESONINTROSPECT': get_meson_script(self.environment, 'mesonintrospect')}
         child_env = os.environ.copy()
         child_env.update(env)
 
         for s in self.build.postconf_scripts:
             cmd = s['exe'] + s['args']
             subprocess.check_call(cmd, env=child_env)
-
-    # Subprojects of subprojects may cause the same dep args to be used
-    # multiple times. Remove duplicates here. Note that we can't dedup
-    # libraries based on name alone, because "-lfoo -lbar -lfoo" is 
-    # a completely valid (though pathological) sequence and removing the
-    # latter may fail. Usually only applies to static libs, though.
-    def dedup_arguments(self, commands):
-        includes = {}
-        final_commands = []
-        previous = '-fsuch_arguments=woof'
-        for c in commands:
-            if c.startswith(('-I', '-L', '/LIBPATH')):
-                if c in includes:
-                    continue
-                includes[c] = True
-            if previous == c:
-                continue
-            previous = c
-            final_commands.append(c)
-        return final_commands
-                                                  
