@@ -65,17 +65,21 @@ def determine_installed_path(target, installdata):
 def list_installed(installdata):
     res = {}
     if installdata is not None:
+        for path, installdir, aliases, unknown1, unknown2 in installdata.targets:
+            res[os.path.join(installdata.build_dir, path)] = os.path.join(installdata.prefix, installdir, os.path.basename(path))
         for path, installpath, unused_prefix in installdata.data:
             res[path] = os.path.join(installdata.prefix, installpath)
+        for path, installdir in installdata.headers:
+            res[path] = os.path.join(installdata.prefix, installdir, os.path.basename(path))
+        for path, installpath in installdata.man:
+            res[path] = os.path.join(installdata.prefix, installpath)
     print(json.dumps(res))
 
 
 def list_targets(coredata, builddata, installdata):
     tlist = []
     for (idname, target) in builddata.get_targets().items():
-        t = {}
-        t['name'] = target.get_basename()
-        t['id'] = idname
+        t = {'name': target.get_basename(), 'id': idname}
         fname = target.get_filename()
         if isinstance(fname, list):
             fname = [os.path.join(target.subdir, x) for x in fname]
@@ -126,9 +130,7 @@ def add_keys(optlist, options):
     keys.sort()
     for key in keys:
         opt = options[key]
-        optdict = {}
-        optdict['name'] = key
-        optdict['value'] = opt.value
+        optdict = {'name': key, 'value': opt.value}
         if isinstance(opt, coredata.UserStringOption):
             typestr = 'string'
         elif isinstance(opt, coredata.UserBooleanOption):
@@ -155,12 +157,12 @@ def list_buildsystem_files(coredata, builddata):
     print(json.dumps(filelist))
 
 def list_deps(coredata):
-    result = {}
-    for d in coredata.deps.values():
+    result = []
+    for d in coredata.deps:
         if d.found():
             args = {'compile_args': d.get_compile_args(),
                     'link_args': d.get_link_args()}
-            result[d.name] = args
+            result += [d.name, args]
     print(json.dumps(result))
 
 def list_tests(testdata):
@@ -184,9 +186,7 @@ def list_tests(testdata):
     print(json.dumps(result))
 
 def list_projinfo(builddata):
-    result = {}
-    result['name'] = builddata.project_name
-    result['version'] = builddata.project_version
+    result = {'name': builddata.project_name, 'version': builddata.project_version}
     subprojects = []
     for k, v in builddata.subprojects.items():
         c = {'name': k,
